from decimal import Decimal
from enum import Enum


class Type(Enum):
    ANY = "any"
    LIST = "list"
    DICT = "dict"
    STRING = "string"
    STRING_OR_NUMBER = "string_or_number"
    DECIMAL = "decimal"
    BOOLEAN = "boolean"
    NUMBER = "number"
    INTEGER = "integer"
    CAMPAIGN = "campaign"
    KEYWORD = "keyword"
    SOV_KEYWORD = "sov_keyword"
    NONE = None


class ValidationError(ValueError):
    pass


class SchemaValidator:
    TOP_LEVEL_NO_PARAM = "__top_level__"
    DEFAULT_SUPPORTED_SCHEMA_PARAMS = ["type", "required"]
    SUPPORTED_TYPES = [
        Type.ANY,
        Type.LIST,
        Type.DICT,
        Type.STRING,
        Type.STRING_OR_NUMBER,
        Type.DECIMAL,
        Type.BOOLEAN,
        Type.NUMBER,
        Type.INTEGER,
        Type.CAMPAIGN,
        Type.KEYWORD,
        Type.SOV_KEYWORD,
    ]
    INSTANCE_TYPE_BY_SCHEMA_TYPE = {
        Type.ANY: tuple(),
        Type.LIST: (list,),
        Type.DICT: (dict,),
        Type.STRING: (str,),
        Type.STRING_OR_NUMBER: (str, Decimal, float, int),
        Type.DECIMAL: (Decimal, float),
        Type.BOOLEAN: (bool,),
        Type.NUMBER: (Decimal, float, int),
        Type.INTEGER: (int,),
        Type.CAMPAIGN: (str,),
        Type.KEYWORD: (str,),
        Type.SOV_KEYWORD: (int,),
    }

    @classmethod
    def supported_schema_params(cls):
        return {
            Type.LIST: [*cls.DEFAULT_SUPPORTED_SCHEMA_PARAMS, "element_schema"],
            Type.DICT: [
                *cls.DEFAULT_SUPPORTED_SCHEMA_PARAMS,
                "schema",
                "arbitrary_keys",
                "conditional_schema_key",
                "conditional_schema",
                "conditional_require_func",
            ],
            Type.STRING: [*cls.DEFAULT_SUPPORTED_SCHEMA_PARAMS, "choices"],
            Type.NUMBER: [*cls.DEFAULT_SUPPORTED_SCHEMA_PARAMS, "min", "max"],
            Type.INTEGER: [*cls.DEFAULT_SUPPORTED_SCHEMA_PARAMS, "min", "max"],
            Type.DECIMAL: [*cls.DEFAULT_SUPPORTED_SCHEMA_PARAMS, "min", "max"],
        }

    @classmethod
    def _get_instance_type_by_schema_type(cls, schema_type):
        schema_type = (
            schema_type if isinstance(schema_type, Type) else Type(schema_type)
        )
        return cls.INSTANCE_TYPE_BY_SCHEMA_TYPE[schema_type]

    @classmethod
    def _get_supported_schema_params_by_schema_type(cls, schema_type):
        schema_type = (
            schema_type if isinstance(schema_type, Type) else Type(schema_type)
        )
        return cls.supported_schema_params().get(
            schema_type, cls.DEFAULT_SUPPORTED_SCHEMA_PARAMS
        )

    @classmethod
    def is_top_level(cls, schema):
        return "schema" in schema

    @classmethod
    def validate_object(cls, obj, schema):
        if schema is None:
            schema = {}

        if cls.is_top_level(schema):
            schema = {cls.TOP_LEVEL_NO_PARAM: schema}
            obj = {cls.TOP_LEVEL_NO_PARAM: obj}

        cls.validate_schema(schema)

        for param_name in schema:
            schema_config = schema[param_name]
            cls._validate_object(param_name, obj, schema_config)

        for param_name in obj:
            if param_name not in schema:
                raise ValidationError(("unsupported field {}".format(param_name)))

    @classmethod
    def _validate_obj_required(cls, param_name, obj, schema_config):
        # Normal case
        required = schema_config.get("required")

        # Conditional case
        if required is None and schema_config.get("conditional_schema_key"):
            # Conditional requirements
            conditional_key_value = obj[schema_config.get("conditional_schema_key")]

            # conditional schemas allow selective required fields
            if schema_config.get("conditional_require_func"):
                required = schema_config.get("conditional_require_func")(
                    conditional_key_value
                )

        # Default to False if neither of the above apply
        if required is None:
            required = False

        if required and param_name not in obj:
            raise ValidationError(("missing required field {}".format(param_name)))

        return required

    @classmethod
    def _validate_object_types(cls, param_name, obj, schema_config):
        schema_type = Type(schema_config.get("type", None))
        required = cls._validate_obj_required(param_name, obj, schema_config)
        acceptable_types = cls._get_instance_type_by_schema_type(schema_type)

        # If the element isn't required, we accept None
        if not required:
            acceptable_types = (*acceptable_types, type(None))

        if schema_type != Type.ANY and not isinstance(
            obj[param_name], acceptable_types
        ):
            raise ValidationError(
                (
                    "wrong type for field {}, must be one of {}".format(
                        param_name, ", ".join([t.__name__ for t in acceptable_types]),
                    )
                )
            )

        if obj[param_name] is not None:
            choices = schema_config.get("choices")
            if choices and obj[param_name] not in choices:
                raise ValidationError(
                    (
                        "wrong value ({}) for field {}, must be one of {}".format(
                            obj[param_name],
                            param_name,
                            ", ".join(schema_config.get("choices")),
                        )
                    )
                )

            if "max" in schema_config and obj[param_name] > schema_config["max"]:
                raise ValidationError(
                    (
                        "value ({}) for field {} is out of range, must be less than {}".format(
                            obj[param_name], param_name, schema_config["max"]
                        )
                    )
                )

            if "min" in schema_config and obj[param_name] < schema_config["min"]:
                raise ValidationError(
                    (
                        "value ({}) for field {} is out of range, must be less than {}".format(
                            obj[param_name], param_name, schema_config["min"]
                        )
                    )
                )

    @classmethod
    def _validate_dict_object(cls, param_name, obj, schema_config):
        if (
            "schema" not in schema_config
            and schema_config.get("conditional_schema_key")
            and obj[schema_config.get("conditional_schema_key")]
            not in schema_config.get("conditional_schema")
        ):
            raise ValidationError(
                (
                    "wrong value for conditional_schema_key referenced field {}, "
                    "must be a valid conditional_schema of {}".format(
                        schema_config.get("conditional_schema_key"),
                        ", ".join(schema_config.get("conditional_schema").keys()),
                    )
                )
            )

        if "schema" in schema_config:
            sub_schema = schema_config.get("schema").copy()
        elif schema_config.get("conditional_schema_key"):
            conditional_key = obj[schema_config.get("conditional_schema_key")]
            sub_schema = schema_config.get("conditional_schema")[conditional_key].copy()
        else:
            return

        sub_obj = obj[param_name]

        # If the schema allows arbitrary keys, we have to preprocess
        if schema_config.get("arbitrary_keys", False):
            arbitrary_key_schema = sub_schema.pop("__arbitrary_keys__", None)
            sub_obj = {}

            for key, val in obj[param_name].items():

                # If we've specified a schema for this entry, or we've specified
                # a schema for arbitrary keys, include it in validation
                if key in sub_schema or arbitrary_key_schema is not None:
                    sub_obj[key] = val

                # If this key doesn't come with its own schema,
                # apply the schema included for arbitrary keys
                if key not in sub_schema and arbitrary_key_schema is not None:
                    sub_schema[key] = arbitrary_key_schema

        if sub_obj or sub_schema:
            cls.validate_object(sub_obj, sub_schema)

    @classmethod
    def _validate_object(cls, param_name, obj, schema_config):
        if param_name in obj:
            cls._validate_object_types(param_name, obj, schema_config)
            schema_type = Type(schema_config.get("type", None))

            if schema_type == Type.LIST:
                for sub_obj in obj[param_name]:
                    sub_name = f"{param_name}__element"
                    cls._validate_object(
                        sub_name, {sub_name: sub_obj}, schema_config["element_schema"]
                    )

            if schema_type == Type.DICT:
                cls._validate_dict_object(param_name, obj, schema_config)

    @classmethod
    def validate_schema(cls, schema):
        for param_name in schema:
            param_config = schema[param_name]
            cls.validate_schema_config(param_name, param_config)

    @classmethod
    def _validate_dict_schema_config(cls, param_name, schema_config):
        if not isinstance(schema_config.get("arbitrary_keys", True), bool):
            raise ValidationError(
                (
                    f"not a valid parameter schema, `{param_name}` "
                    "parameter `arbitrary_keys` must be boolean"
                )
            )

        if schema_config.get("schema"):
            cls.validate_schema(schema_config["schema"])
        elif schema_config.get("conditional_schema") or schema_config.get(
            "conditional_schema_key"
        ):
            if not schema_config.get("conditional_schema_key"):
                raise ValidationError(
                    (
                        f"not a valid parameter schema, `{param_name}` requires "
                        "conditional_schema_key to define which conditional_schema to use"
                    )
                )
            if not schema_config.get("conditional_schema"):
                raise ValidationError(
                    (
                        f"not a valid parameter schema, `{param_name}` "
                        "requires conditional_schema to define which schema "
                        "to use when conditional_schema_key is set"
                    )
                )

            for _, schema in schema_config.get("conditional_schema").items():
                # Permits users to specify no schema for a conditional.
                if schema is not None:
                    cls.validate_schema(schema)
        elif not schema_config.get("arbitrary_keys", False):
            raise ValidationError(
                (
                    f"not a valid parameter schema, `{param_name}` requires a schema "
                    "or conditional_schema and conditional_schema_key to define its contents"
                )
            )

    @classmethod
    def validate_schema_config(cls, param_name, schema_config):
        if not isinstance(schema_config, dict):
            raise ValidationError(
                (
                    f"not a valid parameter schema, `{param_name}` "
                    f"needs dict with containing each of {cls.DEFAULT_SUPPORTED_SCHEMA_PARAMS}"
                )
            )

        schema_type = Type(schema_config.get("type", None))
        if schema_type not in cls.SUPPORTED_TYPES:
            raise ValidationError(
                (
                    f"not a valid parameter schema, `{param_name}` "
                    f"needs string `type` one of `{cls.SUPPORTED_TYPES}`"
                )
            )

        if not isinstance(schema_config.get("required"), bool) and not callable(
            schema_config.get("conditional_require_func")
        ):
            raise ValidationError(
                (
                    f"not a valid parameter schema, `{param_name}` needs boolean `required`"
                )
            )

        supported_schema_params = cls._get_supported_schema_params_by_schema_type(
            schema_type
        )
        unsupported_config_fields = [
            c for c in schema_config if c not in supported_schema_params
        ]
        if unsupported_config_fields:
            raise ValidationError(
                (
                    f"not a valid parameter schema, `{param_name}` has "
                    f"unsupported config parameters `{unsupported_config_fields}`"
                )
            )

        if "choices" in schema_config and schema_type != Type.STRING:
            raise ValidationError(
                (
                    f"not a valid parameter schema, `{param_name}` "
                    "has choices on incompatible schema type, only string supported"
                )
            )

        # For iterables
        if schema_type == Type.LIST:
            if not schema_config.get("element_schema", None):
                raise ValidationError(
                    (
                        f"not a valid parameter schema, `{param_name}` "
                        "requires an element_schema to define its contents"
                    )
                )
            # Validate the nested schema
            cls.validate_schema_config(
                f"{param_name}__element", schema_config["element_schema"]
            )

        # For dicts
        if schema_type == Type.DICT:
            cls._validate_dict_schema_config(param_name, schema_config)
